@@ -22,13 +22,13 @@
 #include <sys/mount.h>
 
 #include <mntent.h>
-#include <paths.h>
 
 #ifdef HAVE_LIBSELINUX
 #include <selinux/selinux.h>
 #include <selinux/context.h>
 #endif
 
+#include "pathnames.h"
 #include "fsprobe.h"
 #include "mount_constants.h"
 #include "sundries.h"
@@ -37,11 +37,11 @@
 #include "fstab.h"
 #include "lomount.h"
 #include "loop.h"
-#include "linux_fs.h"		/* for BLKGETSIZE */
 #include "getusername.h"
-#include "mount_paths.h"
 #include "env.h"
 #include "nls.h"
+#include "blkdev.h"
+#include "realpath.h"
 
 #define DO_PS_FIDDLING
 
@@ -49,9 +49,6 @@
 #include "setproctitle.h"
 #endif
 
-/* Quiet mode */
-int mount_quiet = 0;
-
 /* True for fake mount (-f).  */
 static int fake = 0;
 
@@ -64,17 +61,14 @@ static int nomtab = 0;
 /* True for explicit readonly (-r).  */
 static int readonly = 0;
 
-/* Nonzero for chatty (-v).  */
-int verbose = 0;
-
 /* Nonzero for sloppy (-s).  */
-int sloppy = 0;
+static int sloppy = 0;
 
 /* True for explicit read/write (-w).  */
 static int readwrite = 0;
 
 /* True for all mount (-a).  */
-int mount_all = 0;
+static int mount_all = 0;
 
 /* True for fork() during all mount (-F).  */
 static int optfork = 0;
@@ -82,8 +76,8 @@ static int optfork = 0;
 /* Add volumelabel in a listing of mounted devices (-l). */
 static int list_with_volumelabel = 0;
 
-/* Nonzero for mount {--bind|--replace|--before|--after|--over|--move|
- * 		       make-shared|make-private|make-unbindable|make-slave}
+/* Nonzero for mount {bind|move|make-shared|make-private|
+ *				make-unbindable|make-slave}
  */
 static int mounttype = 0;
 
@@ -182,11 +176,14 @@ static const struct opt_map opt_map[] = {
   { "norelatime", 0, 1, MS_RELATIME }, /* Update access time without regard
 					  to mtime/ctime */
 #endif
+  { "nofail",	0, 0, MS_COMMENT},	/* Do not fail if ENOENT on dev */
   { NULL,	0, 0, 0		}
 };
 
-static const char *opt_loopdev, *opt_vfstype, *opt_offset, *opt_encryption,
-	*opt_speed, *opt_comment, *opt_uhelper;
+static int opt_nofail = 0;
+
+static const char *opt_loopdev, *opt_vfstype, *opt_offset, *opt_sizelimit,
+        *opt_encryption, *opt_speed, *opt_comment, *opt_uhelper;
 
 static int mounted (const char *spec0, const char *node0);
 static int check_special_mountprog(const char *spec, const char *node,
@@ -200,6 +197,7 @@ static struct string_opt_map {
   { "loop=",	0, &opt_loopdev },
   { "vfs=",	1, &opt_vfstype },
   { "offset=",	0, &opt_offset },
+  { "sizelimit=",  0, &opt_sizelimit },
   { "encryption=", 0, &opt_encryption },
   { "speed=", 0, &opt_speed },
   { "comment=", 1, &opt_comment },
@@ -230,12 +228,6 @@ parse_string_opt(char *s) {
 	return 0;
 }
 
-static void
-my_free(const void *s) {
-	if (s)
-		free((void *) s);
-}
-
 /* Report on a single mount.  */
 static void
 print_one (const struct my_mntent *me) {
@@ -389,6 +381,8 @@ parse_opt(char *opt, int *mask, char **extra_opts) {
 				verbose = 0;
 			}
 #endif
+			if (streq(opt, "nofail"))
+				opt_nofail = 1;
 			return;
 		}
 
@@ -529,11 +523,11 @@ create_mtab (void) {
 
 	lock_mtab();
 
-	mfp = my_setmntent (MOUNTED, "a+");
+	mfp = my_setmntent (_PATH_MOUNTED, "a+");
 	if (mfp == NULL || mfp->mntent_fp == NULL) {
 		int errsv = errno;
 		die (EX_FILEIO, _("mount: can't open %s for writing: %s"),
-		     MOUNTED, strerror (errsv));
+		     _PATH_MOUNTED, strerror (errsv));
 	}
 
 	/* Find the root entry by looking it up in fstab */
@@ -545,12 +539,12 @@ create_mtab (void) {
 		mnt.mnt_type = fstab->m.mnt_type;
 		mnt.mnt_opts = fix_opts_string (flags, extra_opts, NULL);
 		mnt.mnt_freq = mnt.mnt_passno = 0;
-		my_free(extra_opts);
+		free(extra_opts);
 
 		if (my_addmntent (mfp, &mnt) == 1) {
 			int errsv = errno;
 			die (EX_FILEIO, _("mount: error writing %s: %s"),
-			     MOUNTED, strerror (errsv));
+			     _PATH_MOUNTED, strerror (errsv));
 		}
 	}
 	if (fchmod (fileno (mfp->mntent_fp), 0644) < 0)
@@ -558,11 +552,13 @@ create_mtab (void) {
 			int errsv = errno;
 			die (EX_FILEIO,
 			     _("mount: error changing mode of %s: %s"),
-			     MOUNTED, strerror (errsv));
+			     _PATH_MOUNTED, strerror (errsv));
 		}
 	my_endmntent (mfp);
 
 	unlock_mtab();
+
+	reset_mtab_info();
 }
 
 /* count successful mount system calls */
@@ -624,7 +620,10 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
   if (!external_allowed)
       return 0;
 
-  if (type && strlen(type) < 100) {
+  if (type == NULL || strcmp(type, "none") == 0)
+	  return 0;
+
+  if (strlen(type) < 100) {
        sprintf(mountprog, "/sbin/mount.%s", type);
        if (stat(mountprog, &statbuf) == 0) {
 	    if (verbose)
@@ -634,8 +633,12 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
 		 char *oo, *mountargs[10];
 		 int i = 0;
 
-		 setuid(getuid());
-		 setgid(getgid());
+		 if(setgid(getgid()) < 0)
+			 die(EX_FAIL, _("mount: cannot set group id: %s"), strerror(errno));
+
+		 if(setuid(getuid()) < 0)
+			 die(EX_FAIL, _("mount: cannot set user id: %s"), strerror(errno));
+
 		 oo = fix_opts_string (flags, extra_opts, NULL);
 		 mountargs[i++] = mountprog;				/* 1 */
 		 mountargs[i++] = (char *) spec;			/* 2 */
@@ -656,7 +659,7 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
 
 		 if (verbose > 2) {
 			i = 0;
-			while(verbose > 2 && mountargs[i]) {
+			while(mountargs[i]) {
 				printf("mount: external mount: argv[%d] = \"%s\"\n",
 					i, mountargs[i]);
 				i++;
@@ -691,8 +694,8 @@ guess_fstype_by_devname(const char *devname)
 
       if (!type)
          printf (_("       I will try all types mentioned in %s or %s\n"),
-		      ETC_FILESYSTEMS, PROC_FILESYSTEMS);
-      else if (!strcmp(type, "swap"))
+		      _PATH_FILESYSTEMS, _PATH_PROC_FILESYSTEMS);
+      else if (!strcmp(type, MNTTYPE_SWAP))
          printf (_("       and it looks like this is swapspace\n"));
       else
          printf (_("       I will try type %s\n"), type);
@@ -715,13 +718,13 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
    if (*types && strcasecmp (*types, "auto") == 0)
       *types = NULL;
 
-   if (!*types && (flags & (MS_BIND | MS_MOVE)))
-      *types = "none";		/* random, but not "bind" */
+   if (flags & (MS_BIND | MS_MOVE))
+      *types = "none";
 
    if (!*types && !(flags & MS_REMOUNT)) {
       *types = guess_fstype_by_devname(spec);
       if (*types) {
-	  if (!strcmp(*types, "swap")) {
+	  if (!strcmp(*types, MNTTYPE_SWAP)) {
 	      error(_("%s looks like swapspace - not mounted"), spec);
 	      *types = NULL;
 	      return 1;
@@ -824,11 +827,53 @@ suid_check(const char *spec, const char *node, int *flags, char **user) {
   *flags &= ~(MS_OWNER | MS_GROUP);
 }
 
+/* Check, if there already exists a mounted loop device on the mountpoint node
+ * with the same parameters.
+ */
+static int
+is_mounted_same_loopfile(const char *node0, const char *loopfile, unsigned long long offset)
+{
+	struct mntentchn *mnt = NULL;
+	char *node;
+	int res = 0;
+
+	node = canonicalize_mountpoint(node0);
+
+	/* Search for mountpoint node in mtab,
+	 * procceed if any of these has the loop option set or
+	 * the device is a loop device
+	 */
+	mnt = getmntdirbackward(node, mnt);
+	if (!mnt) {
+		free(node);
+		return 0;
+	}
+	for(; mnt && res == 0; mnt = getmntdirbackward(node, mnt)) {
+		char *p;
+
+		if (strncmp(mnt->m.mnt_fsname, "/dev/loop", 9) == 0)
+			res = loopfile_used_with((char *) mnt->m.mnt_fsname,
+					loopfile, offset);
+
+		else if ((p = strstr(mnt->m.mnt_opts, "loop="))) {
+			char *dev = xstrdup(p+5);
+			if ((p = strchr(dev, ',')))
+				*p = '\0';
+			res = loopfile_used_with(dev, loopfile, offset);
+			free(dev);
+		}
+	}
+
+	free(node);
+	return res;
+}
+
 static int
 loop_check(const char **spec, const char **type, int *flags,
-	   int *loop, const char **loopdev, const char **loopfile) {
+	   int *loop, const char **loopdev, const char **loopfile,
+	   const char *node) {
   int looptype;
-  unsigned long long offset;
+  unsigned long long offset, sizelimit;
 
   /*
    * In the case of a loop mount, either type is of the form lo@/dev/loop5
@@ -853,7 +898,7 @@ loop_check(const char **spec, const char **type, int *flags,
       *type = opt_vfstype;
   }
 
-  *loop = ((*flags & MS_LOOP) || *loopdev || opt_offset || opt_encryption);
+  *loop = ((*flags & MS_LOOP) || *loopdev || opt_offset || opt_sizelimit || opt_encryption);
   *loopfile = *spec;
 
   if (*loop) {
@@ -862,10 +907,19 @@ loop_check(const char **spec, const char **type, int *flags,
       if (verbose)
 	printf(_("mount: skipping the setup of a loop device\n"));
     } else {
-      int loopro = (*flags & MS_RDONLY);
+      int loop_opts = SETLOOP_AUTOCLEAR; /* always attempt autoclear */
       int res;
 
+      if (*flags & MS_RDONLY)
+        loop_opts |= SETLOOP_RDONLY;
+
       offset = opt_offset ? strtoull(opt_offset, NULL, 0) : 0;
+      sizelimit = opt_sizelimit ? strtoull(opt_sizelimit, NULL, 0) : 0;
+
+      if (is_mounted_same_loopfile(node, *loopfile, offset)) {
+        error(_("mount: according to mtab %s is already mounted on %s as loop"), *loopfile, node);
+        return EX_FAIL;
+      }
 
       do {
         if (!*loopdev || !**loopdev)
@@ -875,8 +929,8 @@ loop_check(const char **spec, const char **type, int *flags,
 	if (verbose)
 	  printf(_("mount: going to use the loop device %s\n"), *loopdev);
 
-	if ((res = set_loop(*loopdev, *loopfile, offset,
-			    opt_encryption, pfd, &loopro))) {
+	if ((res = set_loop(*loopdev, *loopfile, offset, sizelimit,
+			    opt_encryption, pfd, &loop_opts))) {
 	  if (res == 2) {
 	     /* loop dev has been grabbed by some other process,
 	        try again, if not given explicitly */
@@ -905,8 +959,13 @@ loop_check(const char **spec, const char **type, int *flags,
       if (verbose > 1)
 	printf(_("mount: setup loop device successfully\n"));
       *spec = *loopdev;
-      if (loopro)
-	*flags |= MS_RDONLY;
+
+      if (loop_opts & SETLOOP_RDONLY)
+        *flags |= MS_RDONLY;
+
+      if (loop_opts & SETLOOP_AUTOCLEAR)
+        /* Prevent recording loop dev in mtab for cleanup on umount */
+        *loop = 0;
     }
   }
 
@@ -919,7 +978,7 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 	struct my_mntent mnt;
 
 	mnt.mnt_fsname = canonicalize (spec);
-	mnt.mnt_dir = canonicalize (node);
+	mnt.mnt_dir = canonicalize_mountpoint (node);
 	mnt.mnt_type = type;
 	mnt.mnt_opts = opts;
 	mnt.mnt_freq = freq;
@@ -930,6 +989,13 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 	if (verbose)
 		print_one (&mnt);
 
+	if (!nomtab && mtab_does_not_exist()) {
+		if (verbose > 1)
+			printf(_("mount: no %s found - creating it..\n"),
+			       MOUNTED);
+		create_mtab ();
+	}
+
 	if (!nomtab && mtab_is_writable()) {
 		if (flags & MS_REMOUNT)
 			update_mtab (mnt.mnt_dir, &mnt);
@@ -939,16 +1005,16 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 			mntFILE *mfp;
 
 			lock_mtab();
-			mfp = my_setmntent(MOUNTED, "a+");
+			mfp = my_setmntent(_PATH_MOUNTED, "a+");
 			if (mfp == NULL || mfp->mntent_fp == NULL) {
 				int errsv = errno;
-				error(_("mount: can't open %s: %s"), MOUNTED,
+				error(_("mount: can't open %s: %s"), _PATH_MOUNTED,
 				      strerror (errsv));
 			} else {
 				if ((my_addmntent (mfp, &mnt)) == 1) {
 					int errsv = errno;
 					error(_("mount: error writing %s: %s"),
-					      MOUNTED, strerror (errsv));
+					      _PATH_MOUNTED, strerror (errsv));
 				}
 			}
 			my_endmntent(mfp);
@@ -987,16 +1053,14 @@ cdrom_setspeed(const char *spec) {
 
 /*
  * try_mount_one()
- *	Try to mount one file system. When "bg" is 1, this is a retry
- *	in the background. One additional exit code EX_BG is used here.
- *	It is used to instruct the caller to retry the mount in the
- *	background.
+ *	Try to mount one file system.
+ *
  * returns: 0: OK, EX_SYSERR, EX_FAIL, return code from nfsmount,
  *      return status from wait
  */
 static int
 try_mount_one (const char *spec0, const char *node0, const char *types0,
-	       const char *opts0, int freq, int pass, int bg, int ro) {
+	       const char *opts0, int freq, int pass, int ro) {
   int res = 0, status = 0, special = 0;
   int mnt5_res = 0;		/* only for gcc */
   int mnt_err;
@@ -1052,7 +1116,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
        * stale assignments of files to loop devices. Nasty when used for
        * encryption.
        */
-      res = loop_check(&spec, &types, &flags, &loop, &loopdev, &loopfile);
+      res = loop_check(&spec, &types, &flags, &loop, &loopdev, &loopfile, node);
       if (res)
 	  goto out;
   }
@@ -1150,9 +1214,11 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       else if (stat (node, &statbuf))
 	   error (_("mount: mount point %s is a symbolic link to nowhere"),
 		  node);
-      else if (stat (spec, &statbuf))
+      else if (stat (spec, &statbuf)) {
+	   if (opt_nofail)
+		goto out;
 	   error (_("mount: special device %s does not exist"), spec);
-      else {
+      } else {
 	   errno = mnt_err;
 	   perror("mount");
       }
@@ -1160,53 +1226,49 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
     case ENOTDIR:
       if (stat (node, &statbuf) || ! S_ISDIR(statbuf.st_mode))
 	   error (_("mount: mount point %s is not a directory"), node);
-      else if (stat (spec, &statbuf) && errno == ENOTDIR)
+      else if (stat (spec, &statbuf) && errno == ENOTDIR) {
+	   if (opt_nofail)
+              goto out;
 	   error (_("mount: special device %s does not exist\n"
 		    "       (a path prefix is not a directory)\n"), spec);
-      else {
+      } else {
 	   errno = mnt_err;
 	   perror("mount");
       }
       break;
     case EINVAL:
     { int fd;
-      unsigned long size;
-      int warned=0;
+      unsigned long long size = 0;
 
       if (flags & MS_REMOUNT) {
 	error (_("mount: %s not mounted already, or bad option"), node);
       } else {
 	error (_("mount: wrong fs type, bad option, bad superblock on %s,\n"
-	       "       missing codepage or other error"),
+	       "       missing codepage or helper program, or other error"),
 	       spec);
 
-	if (stat(spec, &statbuf) == 0 && S_ISBLK(statbuf.st_mode)
-	   && (fd = open(spec, O_RDONLY | O_NONBLOCK)) >= 0) {
-	  if (ioctl(fd, BLKGETSIZE, &size) == 0) {
-	    if (size == 0 && !loop) {
-	      warned++;
+	if (stat(spec, &statbuf) < 0) {
+	  if (errno == ENOENT)         /* network FS? */
+	    error(_(
+	       "       (for several filesystems (e.g. nfs, cifs) you might\n"
+	       "       need a /sbin/mount.<type> helper program)"));
+
+	} else if (S_ISBLK(statbuf.st_mode)
+	                 && (fd = open(spec, O_RDONLY | O_NONBLOCK)) >= 0) {
+
+	  if (blkdev_get_size(fd, &size) == 0) {
+	    if (size == 0 && !loop)
 	      error(_(
 		 "       (could this be the IDE device where you in fact use\n"
 		 "       ide-scsi so that sr0 or sda or so is needed?)"));
-	    }
-	    if (size && size <= 2) {
-	      warned++;
+
+	    if (size && size <= 2)
 	      error(_(
 		  "       (aren't you trying to mount an extended partition,\n"
 		  "       instead of some logical partition inside?)"));
-	    }
-	  close(fd);
+
+	    close(fd);
 	  }
-#if 0
-	  /* 0xf for SCSI, 0x3f for IDE. One might check /proc/partitions
-	     to see whether this thing really is partitioned.
-	     Do not suggest partitions for /dev/fd0. */
-	  if (!warned && (statbuf.st_rdev & 0xf) == 0) {
-	    warned++;
-	    error ("       (could this be the whole disk device\n"
-		   "       where you need a partition?)");
-	  }
-#endif
 	}
 	error(_(
 		"       In some cases useful info is found in syslog - try\n"
@@ -1254,6 +1316,8 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       break;
     }
     case ENOTBLK:
+      if (opt_nofail)
+        goto out;
       if (stat (spec, &statbuf)) /* strange ... */
 	error (_("mount: %s is not a block device, and stat fails?"), spec);
       else if (S_ISBLK(statbuf.st_mode))
@@ -1266,6 +1330,8 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	error (_("mount: %s is not a block device"), spec);
       break;
     case ENXIO:
+      if (opt_nofail)
+        goto out;
       error (_("mount: %s is not a valid block device"), spec); break;
     case EACCES:  /* pre-linux 1.1.38, 1.1.41 and later */
     case EROFS:   /* linux 1.1.38 and later */
@@ -1292,7 +1358,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	     types = 0;
          error (_("mount: %s%s is write-protected, mounting read-only"),
 		bd, spec0);
-	 res = try_mount_one (spec0, node0, types, opts, freq, pass, bg, 1);
+	 res = try_mount_one (spec0, node0, types, opts, freq, pass, 1);
 	 goto out;
       }
       break;
@@ -1313,19 +1379,6 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   return res;
 }
 
-/*
- * set_proc_name()
- *	Update the argument vector, so that this process may be easily
- *	identified in a "ps" listing.
- */
-static void
-set_proc_name (const char *spec)
-{
-#ifdef DO_PS_FIDDLING
-	setproctitle ("mount", spec);
-#endif
-}
-
 static char *
 subst_string(const char *s, const char *sub, int sublen, const char *repl) {
 	char *n;
@@ -1371,7 +1424,6 @@ is_existing_file (const char *s) {
 static int
 mount_one (const char *spec, const char *node, const char *types,
 	   const char *fstabopts, char *cmdlineopts, int freq, int pass) {
-	int status, status2;
 	const char *nspec;
 	char *opts;
 
@@ -1381,11 +1433,6 @@ mount_one (const char *spec, const char *node, const char *types,
 	/* Merge the fstab and command line options.  */
 	opts = append_opt(opts, cmdlineopts, NULL);
 
-	/* Handle possible LABEL= and UUID= forms of spec */
-	nspec = fsprobe_get_devname_for_mounting(spec);
-	if (nspec)
-		spec = nspec;
-
 	if (types == NULL && !mounttype && !is_existing_file(spec)) {
 		if (strchr (spec, ':') != NULL) {
 			types = "nfs";
@@ -1402,49 +1449,31 @@ mount_one (const char *spec, const char *node, const char *types,
 		}
 	}
 
-	/*
-	 * Try to mount the file system. When the exit status is EX_BG,
-	 * we will retry in the background. Otherwise, we're done.
-	 */
-	status = try_mount_one (spec, node, types, opts, freq, pass, 0, 0);
-	if (status != EX_BG)
-		return status;
+	/* Handle possible LABEL= and UUID= forms of spec */
+	if (types == NULL || (strncmp(types, "nfs", 3) &&
+			      strncmp(types, "cifs", 4) &&
+			      strncmp(types, "smbfs", 5))) {
+		nspec = fsprobe_get_devname_for_mounting(spec);
+		if (nspec)
+			spec = nspec;
+	}
 
-	/*
-	 * Retry in the background.
-	 */
-	printf (_("mount: backgrounding \"%s\"\n"), spec);
-	fflush( stdout );		/* prevent duplicate output */
-	if (fork() > 0)
-		return 0;			/* parent returns "success" */
-	spec = xstrdup(spec);		/* arguments will be destroyed */
-	node = xstrdup(node);		/* by set_proc_name()          */
-	types = xstrdup(types);
-	set_proc_name (spec);		/* make a nice "ps" listing */
-	status2 = try_mount_one (spec, node, types, opts, freq, pass, 1, 0);
-	if (verbose && status2)
-		printf (_("mount: giving up \"%s\"\n"), spec);
-
-	my_free(opts);
-	my_free(node);
-	my_free(types);
-	exit (0);			/* child stops here */
+	return try_mount_one (spec, node, types, opts, freq, pass, 0);
 }
 
 /* Check if an fsname/dir pair was already in the old mtab.  */
 static int
 mounted (const char *spec0, const char *node0) {
 	struct mntentchn *mc, *mc0;
-	char *spec, *node;
+	const char *spec, *node;
 	int ret = 0;
 
 	/* Handle possible UUID= and LABEL= in spec */
-	spec0 = fsprobe_get_devname(spec0);
-	if (!spec0)
+	spec = fsprobe_get_devname(spec0);
+	if (!spec)
 		return ret;
 
-	spec = canonicalize(spec0);
-	node = canonicalize(node0);
+	node = canonicalize_mountpoint(node0);
 
 	mc0 = mtab_head();
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
@@ -1623,10 +1652,6 @@ static struct option longopts[] = {
 	{ "pass-fd", 1, 0, 'p' },
 	{ "types", 1, 0, 't' },
 	{ "bind", 0, 0, 128 },
-	{ "replace", 0, 0, 129 },
-	{ "after", 0, 0, 130 },
-	{ "before", 0, 0, 131 },
-	{ "over", 0, 0, 132 },
 	{ "move", 0, 0, 133 },
 	{ "guess-fstype", 1, 0, 134 },
 	{ "rbind", 0, 0, 135 },
@@ -1680,10 +1705,7 @@ usage (FILE *fp, int n) {
 	  "Other options: [-nfFrsvw] [-o options] [-p passwdfd].\n"
 	  "For many more details, say  man 8 mount .\n"
 	));
-/*
-	  "Union or stack mounts are specified using one of\n"
-	  "       --replace, --after, --before, --over\n"
-*/
+
 	unlock_mtab();
 	exit (n);
 }
@@ -1695,6 +1717,9 @@ getfs(const char *spec, const char *uuid, const char *label)
 	struct mntentchn *mc = NULL;
 	const char *devname = NULL;
 
+	if (!spec && !uuid && !label)
+		return NULL;
+
 	/*
 	 * A) 99% of all cases, the spec on cmdline matches
 	 *    with spec in fstab
@@ -1757,15 +1782,28 @@ getfs(const char *spec, const char *uuid, const char *label)
 	 *    Earlier mtab was tried first, but this would sometimes try the
 	 *    wrong mount in case mtab had the root device entry wrong.
 	 */
-	if (!mc)
+	if (!mc && (devname || spec))
 		mc = getmntfile (devname ? devname : spec);
 
-	if (devname)
-		my_free(devname);
+	my_free(devname);
 	return mc;
 }
 
-char *progname;
+
+static void
+print_version(int rc) {
+	printf(	"mount from %s (with "
+#ifdef HAVE_LIBBLKID
+		"libblkid"
+#else
+		"libvolume_id"
+#endif
+#ifdef HAVE_LIBSELINUX
+		" and selinux"
+#endif
+		" support)\n", PACKAGE_STRING);
+	exit(rc);
+}
 
 int
 main(int argc, char *argv[]) {
@@ -1856,8 +1894,8 @@ main(int argc, char *argv[]) {
 			++verbose;
 			break;
 		case 'V':		/* version */
-			printf ("mount (%s)\n", PACKAGE_STRING);
-			exit (0);
+			print_version(EXIT_SUCCESS);
+			break;
 		case 'w':		/* mount read/write */
 			readwrite = 1;
 			readonly = 0;
@@ -1868,18 +1906,6 @@ main(int argc, char *argv[]) {
 		case 128: /* bind */
 			mounttype = MS_BIND;
 			break;
-		case 129: /* replace */
-			mounttype = MS_REPLACE;
-			break;
-		case 130: /* after */
-			mounttype = MS_AFTER;
-			break;
-		case 131: /* before */
-			mounttype = MS_BEFORE;
-			break;
-		case 132: /* over */
-			mounttype = MS_OVER;
-			break;
 		case 133: /* move */
 			mounttype = MS_MOVE;
 			break;
@@ -1937,9 +1963,10 @@ main(int argc, char *argv[]) {
 	}
 
 	if (verbose > 2) {
-		printf("mount: fstab path: \"%s\"\n", _PATH_FSTAB);
-		printf("mount: lock path:  \"%s\"\n", MOUNTED_LOCK);
-		printf("mount: temp path:  \"%s\"\n", MOUNTED_TEMP);
+		printf("mount: fstab path: \"%s\"\n", _PATH_MNTTAB);
+		printf("mount: mtab path:  \"%s\"\n", _PATH_MOUNTED);
+		printf("mount: lock path:  \"%s\"\n", _PATH_MOUNTED_LOCK);
+		printf("mount: temp path:  \"%s\"\n", _PATH_MOUNTED_TMP);
 	}
 
 	argc -= optind;
@@ -1960,12 +1987,7 @@ main(int argc, char *argv[]) {
 			die (EX_USAGE, _("mount: only root can do that"));
 	}
 
-	if (!nomtab && mtab_does_not_exist()) {
-		if (verbose > 1)
-			printf(_("mount: no %s found - creating it..\n"),
-			       MOUNTED);
-		create_mtab ();
-	}
+	atexit(unlock_mtab);
 
 	switch (argc+specseen) {
 	case 0:
@@ -1977,19 +1999,24 @@ main(int argc, char *argv[]) {
 
 	case 1:
 		/* mount [-nfrvw] [-o options] special | node
+		 * mount -L label  (or -U uuid)
 		 * (/etc/fstab is necessary)
 		 */
 		if (types != NULL)
 			usage (stderr, EX_USAGE);
 
-		mc = getfs(*argv, uuid, label);
+		if (uuid || label)
+			mc = getfs(NULL, uuid, label);
+		else
+			mc = getfs(*argv, NULL, NULL);
+
 		if (!mc) {
 			if (uuid || label)
 				die (EX_USAGE, _("mount: no such partition found"));
 
 			die (EX_USAGE,
 			     _("mount: can't find %s in %s or %s"),
-			     *argv, _PATH_FSTAB, MOUNTED);
+			     *argv, _PATH_MNTTAB, _PATH_MOUNTED);
 		}
 
 		result = mount_one (xstrdup (mc->m.mnt_fsname),
